// WL_DRAW.C

#include "WL_DEF.H"
#include <DOS.H>
#pragma hdrstop

//#define DEBUGWALLS
//#define DEBUGTICS

/*
=============================================================================

						 LOCAL CONSTANTS

=============================================================================
*/

// the door is the last picture before the sprites
#define DOORWALL	(PMSpriteStart-14)

#define ACTORSIZE	0x4000

/*
=============================================================================

						 GLOBAL VARIABLES

=============================================================================
*/

extern byte far shadetable[SHADE_COUNT][256];
extern int LSHADE_flag; 

#ifdef DEBUGWALLS
unsigned screenloc[3]= {0,0,0};
#else
unsigned screenloc[3]= {PAGE1START,PAGE2START,PAGE3START};
#endif
unsigned freelatch = FREESTART;

long 	lasttimecount;
long 	frameon;

unsigned	wallheight[MAXVIEWWIDTH], min_wallheight;

fixed	tileglobal	= TILEGLOBAL;
fixed	mindist		= MINDIST;


//
// math tables
//
int			pixelangle[MAXVIEWWIDTH];
long		far finetangent[FINEANGLES/4];
fixed 		far sintable[ANGLES+ANGLES/4],far *costable = sintable+(ANGLES/4);

//
// refresh variables
//
fixed	viewx,viewy;			// the focal point
int		viewangle;
fixed	viewsin,viewcos;



fixed	FixedByFrac (fixed a, fixed b);
void	TransformActor (objtype *ob);
void	BuildTables (void);
void	ClearScreen (void);
int		CalcRotate (objtype *ob);
void	DrawScaleds (void);
void	CalcTics (void);
void	FixOfs (void);
void	ThreeDRefresh (void);



//
// wall optimization variables
//
int		lastside;		// true for vertical
long	lastintercept;
int		lasttilehit;


//
// ray tracing variables
//
int			focaltx,focalty,viewtx,viewty;

int			midangle,angle;
unsigned	xpartial,ypartial;
unsigned	xpartialup,xpartialdown,ypartialup,ypartialdown;
unsigned	xinttile,yinttile;

unsigned	tilehit;
unsigned	pixx;

int		xtile,ytile;
int		xtilestep,ytilestep;
long	xintercept,yintercept;
long	xstep,ystep;

int		horizwall[MAXWALLTILES],vertwall[MAXWALLTILES];


/*
=============================================================================

						 LOCAL VARIABLES

=============================================================================
*/


void AsmRefresh (void);			// in WL_DR_A.ASM

/*
============================================================================

			   3 - D  DEFINITIONS

============================================================================
*/


//==========================================================================


/*
========================
=
= FixedByFrac
=
= multiply a 16/16 bit, 2's complement fixed point number by a 16 bit
= fraction, passed as a signed magnitude 32 bit number
=
========================
*/

#pragma warn -rvl			// I stick the return value in with ASMs

fixed FixedByFrac (fixed a, fixed b)
{
//
// setup
//
asm	mov	si,[WORD PTR b+2]	// sign of result = sign of fraction

asm	mov	ax,[WORD PTR a]
asm	mov	cx,[WORD PTR a+2]

asm	or	cx,cx
asm	jns	aok:				// negative?
asm	neg	cx
asm	neg	ax
asm	sbb	cx,0
asm	xor	si,0x8000			// toggle sign of result
aok:

//
// multiply  cx:ax by bx
//
asm	mov	bx,[WORD PTR b]
asm	mul	bx					// fraction*fraction
asm	mov	di,dx				// di is low word of result
asm	mov	ax,cx				//
asm	mul	bx					// units*fraction
asm add	ax,di
asm	adc	dx,0

//
// put result dx:ax in 2's complement
//
asm	test	si,0x8000		// is the result negative?
asm	jz	ansok:
asm	neg	dx
asm	neg	ax
asm	sbb	dx,0

ansok:;

}

#pragma warn +rvl

//==========================================================================

/*
========================
=
= TransformActor
=
= Takes paramaters:
=   gx,gy		: globalx/globaly of point
=
= globals:
=   viewx,viewy		: point of view
=   viewcos,viewsin	: sin/cos of viewangle
=   scale		: conversion from global value to screen value
=
= sets:
=   screenx,transx,transy,screenheight: projected edge location and size
=
========================
*/


//
// transform actor
//
void TransformActor (objtype *ob)
{
	int ratio;
	fixed gx,gy,gxt,gyt,nx,ny;
	long	temp;

//
// translate point to view centered coordinates
//
	gx = ob->x-viewx;
	gy = ob->y-viewy;

//
// calculate newx
//
	gxt = FixedByFrac(gx,viewcos);
	gyt = FixedByFrac(gy,viewsin);
	nx = gxt-gyt-ACTORSIZE;		// fudge the shape forward a bit, because
								// the midpoint could put parts of the shape
								// into an adjacent wall

//
// calculate newy
//
	gxt = FixedByFrac(gx,viewsin);
	gyt = FixedByFrac(gy,viewcos);
	ny = gyt+gxt;

//
// calculate perspective ratio
//
	ob->transx = nx;
	ob->transy = ny;

	if (nx<mindist)			// too close, don't overflow the divide
	{
	  ob->viewheight = 0;
	  return;
	}

	ob->viewx = centerx + ny*scale/nx;	// DEBUG: use assembly divide

//
// calculate height (heightnumerator/(nx>>8))
//
	asm	mov	ax,[WORD PTR heightnumerator]
	asm	mov	dx,[WORD PTR heightnumerator+2]
	asm	idiv	[WORD PTR nx+1]			// nx>>8
	asm	mov	[WORD PTR temp],ax
	asm	mov	[WORD PTR temp+2],dx

	ob->viewheight = temp;
}

//==========================================================================

/*
========================
=
= TransformTile
=
= Takes paramaters:
=   tx,ty		: tile the object is centered in
=
= globals:
=   viewx,viewy		: point of view
=   viewcos,viewsin	: sin/cos of viewangle
=   scale		: conversion from global value to screen value
=
= sets:
=   screenx,transx,transy,screenheight: projected edge location and size
=
= Returns true if the tile is withing getting distance
=
========================
*/

boolean TransformTile (int tx, int ty, int *dispx, int *dispheight)
{
	int ratio;
	fixed gx,gy,gxt,gyt,nx,ny;
	long	temp;

//
// translate point to view centered coordinates
//
	gx = ((long)tx<<TILESHIFT)+0x8000-viewx;
	gy = ((long)ty<<TILESHIFT)+0x8000-viewy;

//
// calculate newx
//
	gxt = FixedByFrac(gx,viewcos);
	gyt = FixedByFrac(gy,viewsin);
	nx = gxt-gyt-0x2000;		// 0x2000 is size of object

//
// calculate newy
//
	gxt = FixedByFrac(gx,viewsin);
	gyt = FixedByFrac(gy,viewcos);
	ny = gyt+gxt;


//
// calculate perspective ratio
//
	if (nx<mindist)			// too close, don't overflow the divide
	{
		*dispheight = 0;
		return false;
	}

	*dispx = centerx + ny*scale/nx;	// DEBUG: use assembly divide

//
// calculate height (heightnumerator/(nx>>8))
//
	asm	mov	ax,[WORD PTR heightnumerator]
	asm	mov	dx,[WORD PTR heightnumerator+2]
	asm	idiv	[WORD PTR nx+1]			// nx>>8
	asm	mov	[WORD PTR temp],ax
	asm	mov	[WORD PTR temp+2],dx

	*dispheight = temp;

//
// see if it should be grabbed
//
	if (nx<TILEGLOBAL && ny>-TILEGLOBAL/2 && ny<TILEGLOBAL/2)
		return true;
	else
		return false;
}

//==========================================================================

/*
====================
=
= CalcHeight
=
= Calculates the height of xintercept,yintercept from viewx,viewy
=
====================
*/

#pragma warn -rvl			// I stick the return value in with ASMs

int	CalcHeight (void)
{
	int	transheight;
	int ratio;
	fixed gxt,gyt,nx,ny;
	long	gx,gy;

	gx = xintercept-viewx;
	gxt = FixedByFrac(gx,viewcos);

	gy = yintercept-viewy;
	gyt = FixedByFrac(gy,viewsin);

	nx = gxt-gyt;

  //
  // calculate perspective ratio (heightnumerator/(nx>>8))
  //
	if (nx<mindist)
		nx=mindist;			// don't let divide overflow

	asm	mov	ax,[WORD PTR heightnumerator]
	asm	mov	dx,[WORD PTR heightnumerator+2]
	asm	idiv	[WORD PTR nx+1]			// nx>>8

	asm mov bx,[min_wallheight]
	asm cmp ax,bx
	asm jae noupdate:
	asm mov [min_wallheight],ax
	noupdate:; 

}


//==========================================================================

/*
===================
=
= ScalePost
=
===================
*/

long		postsource;
unsigned	postx;
unsigned	postwidth;

extern void near ScalePost(); 

void  FarScalePost (void)				// just so other files can call
{
	ScalePost ();
}


/*
====================
=
= HitVertWall
=
= tilehit bit 7 is 0, because it's not a door tile
= if bit 6 is 1 and the adjacent tile is a door tile, use door side pic
=
====================
*/

void HitVertWall (void)
{
	int			wallpic;
	unsigned	texture;

	texture = (yintercept>>4)&0xfc0;
	if (xtilestep == -1)
	{
		texture = 0xfc0-texture;
		xintercept += TILEGLOBAL;
	}
	wallheight[pixx] = CalcHeight();

    	switch(*(mapsegs[1]+farmapylookup[ytile]+xtile))
    	{
    	case ANIMTILE:
          	          tilehit+=(frameon>>6)%4;
          	          break;
    	case ANIMTILE+1:
          	          tilehit+=(frameon>>5)%4;
          	          break;
    	case ANIMTILE+2:
          	          tilehit+=(frameon>>4)%4;
          	          break;
    	case ANIMTILE+3:
          	          tilehit+=(frameon>>3)%4;
          	          break;
    	case ANIMTILE+4:
          	          tilehit+=(frameon>>2)%4;
          	          break;
    	}

	if (lastside==1 && lastintercept == xtile && lasttilehit == tilehit)
	{
		// in the same wall type as last time, so check for optimized draw
		if (texture == (unsigned)postsource && postwidth<8)
		{
		// wide scale
			postwidth++;
			wallheight[pixx] = wallheight[pixx-1];
			return;
		}
		else
		{
			ScalePost ();
			(unsigned)postsource = texture;
			postwidth = 1;
			postx = pixx;
		}
	}
	else
	{
	// new wall
		if (lastside != -1)				// if not the first scaled post
			ScalePost ();

		lastside = true;
		lastintercept = xtile;

		lasttilehit = tilehit;
		postx = pixx;
		postwidth = 1;

		if (tilehit & 0x40)
		{								// check for adjacent doors
			ytile = yintercept>>TILESHIFT;
			if ( tilemap[xtile-xtilestep][ytile]&0x80 )
				wallpic = DOORWALL+9;
			else
				wallpic = vertwall[tilehit & ~0x40];
		}
		else
			wallpic = vertwall[tilehit];

		*( ((unsigned *)&postsource)+1) = (unsigned)PM_GetPage(wallpic);
		(unsigned)postsource = texture;

	}
}


/*
====================
=
= HitHorizWall
=
= tilehit bit 7 is 0, because it's not a door tile
= if bit 6 is 1 and the adjacent tile is a door tile, use door side pic
=
====================
*/

void HitHorizWall (void)
{
	int			wallpic;
	unsigned	texture;

	texture = (xintercept>>4)&0xfc0;
	if (ytilestep == -1)
		yintercept += TILEGLOBAL;
	else
		texture = 0xfc0-texture;
	wallheight[pixx] = CalcHeight();

    	switch(*(mapsegs[1]+farmapylookup[ytile]+xtile))
    	{
    	case ANIMTILE:
          	          tilehit+=(frameon>>6)%4;
          	          break;
    	case ANIMTILE+1:
          	          tilehit+=(frameon>>5)%4;
          	          break;
    	case ANIMTILE+2:
          	          tilehit+=(frameon>>4)%4;
          	          break;
    	case ANIMTILE+3:
          	          tilehit+=(frameon>>3)%4;
          	          break;
    	case ANIMTILE+4:
          	          tilehit+=(frameon>>2)%4;
          	          break;
    	}

	if (lastside==0 && lastintercept == ytile && lasttilehit == tilehit)
	{
		// in the same wall type as last time, so check for optimized draw
		if (texture == (unsigned)postsource && postwidth<8) 
		{
		// wide scale
			postwidth++;
			wallheight[pixx] = wallheight[pixx-1];
			return;
		}
		else
		{
			ScalePost ();
			(unsigned)postsource = texture;
			postwidth = 1;
			postx = pixx;
		}
	}
	else
	{
	// new wall
		if (lastside != -1)				// if not the first scaled post
			ScalePost ();

		lastside = 0;
		lastintercept = ytile;

		lasttilehit = tilehit;
		postx = pixx;
		postwidth = 1;

		if (tilehit & 0x40)
		{								// check for adjacent doors
			xtile = xintercept>>TILESHIFT;
			if ( tilemap[xtile][ytile-ytilestep]&0x80 )
				wallpic = DOORWALL+8;
			else
				wallpic = horizwall[tilehit & ~0x40];
		}
		else
			wallpic = horizwall[tilehit];

		*( ((unsigned *)&postsource)+1) = (unsigned)PM_GetPage(wallpic);
		(unsigned)postsource = texture;
	}

}

//==========================================================================

/*
====================
=
= HitHorizDoor
=
====================
*/

void HitHorizDoor (void)
{
	unsigned	texture,doorpage,doornum;

	doornum = tilehit&0x7f;
	texture = ( (xintercept-doorposition[doornum]) >> 4) &0xfc0;

	wallheight[pixx] = CalcHeight();

	if (lasttilehit == tilehit)
	{
	// in the same door as last time, so check for optimized draw
		if (texture == (unsigned)postsource && postwidth<8)
		{
		// wide scale
			postwidth++;
			wallheight[pixx] = wallheight[pixx-1];
			return;
		}
		else
		{
			ScalePost ();
			(unsigned)postsource = texture;
			postwidth = 1;
			postx = pixx;
		}
	}
	else
	{
		if (lastside != -1)				// if not the first scaled post
			ScalePost ();			// draw last post
	// first pixel in this door
		lastside = 2;
		lasttilehit = tilehit;
		postx = pixx;
		postwidth = 1;

		switch (doorobjlist[doornum].lock)
		{
		case dr_normal:
			doorpage = DOORWALL+6;
			break;
		case dr_lock1:
			doorpage = DOORWALL+4;
			break;
		case dr_lock2:
			doorpage = DOORWALL+12;
			break;		
		case dr_lock3:
			doorpage = DOORWALL+2;
			break;		
		case dr_lock4:
			doorpage = DOORWALL;
			break;
		case dr_elevator:
			doorpage = DOORWALL+10;
			break;
		}

		*( ((unsigned *)&postsource)+1) = (unsigned)PM_GetPage(doorpage);
		(unsigned)postsource = texture;
	}
}

//==========================================================================

/*
====================
=
= HitVertDoor
=
====================
*/

void HitVertDoor (void)
{
	unsigned	texture,doorpage,doornum;

	doornum = tilehit&0x7f;
	texture = ( (yintercept-doorposition[doornum]) >> 4) &0xfc0;

	wallheight[pixx] = CalcHeight();

	if (lasttilehit == tilehit)
	{
	// in the same door as last time, so check for optimized draw
		if (texture == (unsigned)postsource && postwidth<8) 
		{
		// wide scale
			postwidth++;
			wallheight[pixx] = wallheight[pixx-1];
			return;
		}
		else
		{
			ScalePost ();
			(unsigned)postsource = texture;
			postwidth = 1;
			postx = pixx;
		}
	}
	else
	{
		if (lastside != -1)				// if not the first scaled post
			ScalePost ();			// draw last post
	// first pixel in this door
		lastside = 2;
		lasttilehit = tilehit;
		postx = pixx;
		postwidth = 1;

		switch (doorobjlist[doornum].lock)
		{
		case dr_normal:
			doorpage = DOORWALL+6;
			break;
		case dr_lock1:
			doorpage = DOORWALL+4;
			break;
		case dr_lock2:
			doorpage = DOORWALL+12;
			break;		
		case dr_lock3:
			doorpage = DOORWALL+2;
			break;		
		case dr_lock4:
			doorpage = DOORWALL;
			break;
		case dr_elevator:
			doorpage = DOORWALL+10;
			break;
		}

		*( ((unsigned *)&postsource)+1) = (unsigned)PM_GetPage(doorpage+1);
		(unsigned)postsource = texture;
	}
}

//==========================================================================


/*
====================
=
= HitHorizPWall
=
= A pushable wall in action has been hit
=
====================
*/

void HitHorizPWall (void)
{
	int			wallpic;
	unsigned	texture,offset;

	texture = (xintercept>>4)&0xfc0;
	offset = pwallpos<<10;
	if (ytilestep == -1)
		yintercept += TILEGLOBAL-offset;
	else
	{
		texture = 0xfc0-texture;
		yintercept += offset;
	}

	wallheight[pixx] = CalcHeight();

	if (lasttilehit == tilehit)
	{
		// in the same wall type as last time, so check for optimized draw
		if (texture == (unsigned)postsource && postwidth<8) 
		{
		// wide scale
			postwidth++;
			wallheight[pixx] = wallheight[pixx-1];
			return;
		}
		else
		{
			ScalePost ();
			(unsigned)postsource = texture;
			postwidth = 1;
			postx = pixx;
		}
	}
	else
	{
	// new wall
		if (lastside != -1)				// if not the first scaled post
			ScalePost ();

		lasttilehit = tilehit;
		postx = pixx;
		postwidth = 1;

		wallpic = horizwall[tilehit&63];

		*( ((unsigned *)&postsource)+1) = (unsigned)PM_GetPage(wallpic);
		(unsigned)postsource = texture;
	}

}


/*
====================
=
= HitVertPWall
=
= A pushable wall in action has been hit
=
====================
*/

void HitVertPWall (void)
{
	int			wallpic;
	unsigned	texture,offset;

	texture = (yintercept>>4)&0xfc0;
	offset = pwallpos<<10;
	if (xtilestep == -1)
	{
		xintercept += TILEGLOBAL-offset;
		texture = 0xfc0-texture;
	}
	else
		xintercept += offset;

	wallheight[pixx] = CalcHeight();

	if (lasttilehit == tilehit)
	{
		// in the same wall type as last time, so check for optimized draw
		if (texture == (unsigned)postsource && postwidth<8) 
		{
		// wide scale
			postwidth++;
			wallheight[pixx] = wallheight[pixx-1];
			return;
		}
		else
		{
			ScalePost ();
			(unsigned)postsource = texture;
			postwidth = 1;
			postx = pixx;
		}
	}
	else
	{
	// new wall
		if (lastside != -1)				// if not the first scaled post
			ScalePost ();

		lasttilehit = tilehit;
		postx = pixx;
		postwidth = 1;

		wallpic = vertwall[tilehit&63];

		*( ((unsigned *)&postsource)+1) = (unsigned)PM_GetPage(wallpic);
		(unsigned)postsource = texture;
	}

}

//==========================================================================

//==========================================================================

#if 0
/*
=====================
=
= ClearScreen
=
=====================
*/

void ClearScreen (void)
{
 unsigned floor=egaFloor[gamestate.episode*10+mapon],
	  ceiling=egaCeiling[gamestate.episode*10+mapon];

  //
  // clear the screen
  //
asm	mov	dx,GC_INDEX
asm	mov	ax,GC_MODE + 256*2		// read mode 0, write mode 2
asm	out	dx,ax
asm	mov	ax,GC_BITMASK + 255*256
asm	out	dx,ax

asm	mov	dx,40
asm	mov	ax,[viewwidth]
asm	shr	ax,3
asm	sub	dx,ax					// dx = 40-viewwidth/8

asm	mov	bx,[viewwidth]
asm	shr	bx,4					// bl = viewwidth/16
asm	mov	bh,BYTE PTR [viewheight]
asm	shr	bh,1					// half height

asm	mov	ax,[ceiling]
asm	mov	es,[screenseg]
asm	mov	di,[bufferofs]

toploop:
asm	mov	cl,bl
asm	rep	stosw
asm	add	di,dx
asm	dec	bh
asm	jnz	toploop

asm	mov	bh,BYTE PTR [viewheight]
asm	shr	bh,1					// half height
asm	mov	ax,[floor]

bottomloop:
asm	mov	cl,bl
asm	rep	stosw
asm	add	di,dx
asm	dec	bh
asm	jnz	bottomloop


asm	mov	dx,GC_INDEX
asm	mov	ax,GC_MODE + 256*10		// read mode 1, write mode 2
asm	out	dx,ax
asm	mov	al,GC_BITMASK
asm	out	dx,al

}
#endif
//==========================================================================

/*
=====================
=
= SetPixel
=
=====================
*/

 void SetPixel(int x,int y,byte col)
{
   asm{
      cli
      mov al,2
      mov cl,byte ptr x
      and cl,3
      mov ah,1
      shl ah,cl
      mov dx,0x3c4
      out dx,ax
      sti
      push es
      push di
      mov ax,0xa000
      mov es,ax
      mov di,word ptr bufferofs
      mov bx,word ptr y
      mov ax,word ptr x
      shr ax,2
      add di,ax
      shl bx,4
      add di,bx
      shl bx,2
      add di,bx
      mov al,byte ptr col
      mov es:[di],al
      pop di
      pop es
   }
}

/*
=====================
=
= ClearScreenH
=
=====================
*/

void ClearScreenH()
{
   asm{
      cli
      mov al,2
      mov ah,15
      shl ah,cl
      mov dx,0x3c4
      out dx,ax
      sti
      push es
      push di
      mov ax,0xa000
      mov es,ax
      mov di,word ptr bufferofs
      mov bx,320
      mov dx,word ptr viewwidth
      sub bx,dx
      shr dx,3
      shr bx,2
      mov dh,byte ptr viewheight
      shr dh,1
      xor ax,ax
   }
nextrow:
   asm{
      mov cl,dl
      rep stosw
      add di,bx
      dec dh
      jne nextrow
      pop di
      pop es
   }
}

short far starsx[MAXSTARS];
short far starsy[MAXSTARS];
short far starsz[MAXSTARS];

byte moon[100]={
     0,  0, 27, 18, 15, 16, 19, 29,  0,  0,
     0, 22, 16, 15, 15, 16, 16, 18, 24,  0,
    27, 17, 15, 17, 16, 16, 17, 17, 18, 29,
    18, 15, 15, 15, 16, 16, 17, 17, 18, 20,
    16, 15, 15, 16, 16, 17, 17, 18, 19, 21,
    16, 15, 17, 20, 18, 17, 18, 18, 20, 22,
    19, 16, 18, 19, 17, 17, 18, 19, 22, 24,
    28, 19, 17, 17, 17, 18, 19, 21, 25, 31,
     0, 23, 18, 19, 18, 20, 22, 24, 28,  0,
     0,  0, 28, 21, 20, 22, 28, 30,  0,  0 };

/*
=====================
=
= PaintStars
=
=====================
*/

void PaintStars(long siny,long cosy)
{
   long x,y,z;
   int xx,yy;
   int i;
   int shade;
   int hvheight=viewheight>>1;
   int hvwidth=viewwidth>>1;
   ClearScreenH();
   for(i=0;i<MAXSTARS;i++)
   {
      x=(long)starsx[i]*cosy+(long)starsz[i]*siny;
      y=(long)starsy[i]<<16;
      z=((long)starsz[i]*cosy-(long)starsx[i]*siny)>>8;
      if(z<=0) continue;
      shade=(z>>18);
      if(shade>15) continue;
      xx=(int)(x/z+hvwidth);
      yy=(int)(hvheight-y/z);
      if(xx>=0 && xx<viewwidth && yy>=0 && yy<hvheight)
         SetPixel(xx,yy,shade+15);
   }

   x=16384L*cosy+16384L*siny;
   z=(16384L*cosy-16384L*siny)>>8;
   if(z<=0) return;
   xx=(int)(x/z+hvwidth);
   yy=(int)(hvheight-(((long)hvheight-((long)hvheight>>3))<<22)/z);
   if(xx>-10 && xx<viewwidth)
   {
      int stop=10,j;
      i=0;
      if(xx<0) i=-xx;
      if(xx>viewwidth-11) stop=viewwidth-xx;
      for(;i<stop;i++)
         for(j=0;j<10;j++)
            SetPixel(xx+i,yy+j,moon[j*10+i]);
   }
}

word rainpos=0;

/*
=====================
=
= PaintRain
=
=====================
*/

void PaintRain(long siny,long cosy)
{
   long px=(player->y+FixedByFrac(0x7900,siny))>>6;
   long pz=(player->x-FixedByFrac(0x7900,cosy))>>6;
   long ax,az;
   long x,y,z;
   int xx,yy,height,actheight;
   int i;
   int shade;
   int hvheight=viewheight>>1;
   int hvwidth=viewwidth>>1;

   if(siny<0) siny=-(siny&0x7fffffff);
   if(cosy<0) cosy=-(cosy&0x7fffffff);

   rainpos-=1800;
   for(i=0;i<MAXSTARS;i++)
   {
      ax=(long)starsx[i]+px;
      ax=0x1fff-(ax&0x3fff);
      az=(long)starsz[i]+pz;
      az=0x1fff-(az&0x3fff);
      x=ax*cosy+az*siny;
      y=-(heightnumerator<<7)
        +((((long)(starsy[i]<<6)+(long)rainpos)&0x0ffff)<<11);
      z=(az*cosy-ax*siny)>>8;
      if(z<=0) continue;
      shade=(z>>17);
      if(shade>13) continue;
      xx=(int)(x/z+hvwidth);
      if(xx<0 || xx>viewwidth) continue;
      actheight=(int)(y/z);
      yy=hvheight-actheight;
      height=(int)((heightnumerator<<10)/z);
      if(actheight<0) actheight=-actheight;
      if(actheight<(wallheight[xx]>>3) && height<wallheight[xx]) continue;
      if(xx>=0 && xx<viewwidth && yy>0 && yy<viewheight)
      {

         SetPixel(xx,yy,shade+15);
         SetPixel(xx,yy-1,shade+16);
         if(yy>2)
         {
            SetPixel(xx,yy-2,shade+17);
         }
      }
   }
}

/*
=====================
=
= PaintSnow
=
=====================
*/

void PaintSnow(long siny,long cosy)
{
   long px=(player->y+FixedByFrac(0x7900,siny))>>6;
   long pz=(player->x-FixedByFrac(0x7900,cosy))>>6;
   long ax,az;
   long x,y,z;
   int xx,yy,height,actheight;
   int i;
   int shade;
   int hvheight=viewheight>>1;
   int hvwidth=viewwidth>>1;

   if(siny<0) siny=-(siny&0x7fffffff);
   if(cosy<0) cosy=-(cosy&0x7fffffff);

   rainpos-=512;
   for(i=0;i<MAXSTARS;i++)
   {
      ax=(long)starsx[i]+px;
      ax=0x1fff-(ax&0x3fff);
      az=(long)starsz[i]+pz;
      az=0x1fff-(az&0x3fff);
      x=ax*cosy+az*siny;
      y=-(heightnumerator<<7)
        +((((long)(starsy[i]<<6)+(long)rainpos)&0x0ffff)<<11);
      z=(az*cosy-ax*siny)>>8;
      if(z<=0) continue;
      shade=(z>>17);
      if(shade>13) continue;
      xx=(int)(x/z+hvwidth);
      if(xx<0 || xx>viewwidth) continue;
      actheight=(int)(y/z);
      yy=hvheight-actheight;
      height=(int)((heightnumerator<<10)/z);
      if(actheight<0) actheight=-actheight;
      if(actheight<(wallheight[xx]>>3) && height<wallheight[xx]) continue;
      if(xx>0 && xx<viewwidth && yy>0 && yy<viewheight)
      {
         if(shade<10)
         {

            SetPixel(xx,yy,shade+17);
            SetPixel(xx-1,yy,shade+16);
            SetPixel(xx,yy-1,shade+16);
            SetPixel(xx-1,yy-1,shade+15);
         }
         else SetPixel(xx,yy,shade+15);
      }
   }
}

unsigned vgaCeiling[]=
{
#ifndef SPEAR
 0x1d1d,0x1d1d,0x1d1d,0x1d1d,0x1d1d,0x1d1d,0x1d1d,0x1d1d,0x1d1d,0xbfbf,
 0x4e4e,0x4e4e,0x4e4e,0x1d1d,0x8d8d,0x4e4e,0x1d1d,0x2d2d,0x1d1d,0x8d8d,
 0x1d1d,0x1d1d,0x1d1d,0x1d1d,0x1d1d,0x2d2d,0xdddd,0x1d1d,0x1d1d,0x9898,

 0x1d1d,0x9d9d,0x2d2d,0xdddd,0xdddd,0x9d9d,0x2d2d,0x4d4d,0x1d1d,0xdddd,
 0x7d7d,0x1d1d,0x2d2d,0x2d2d,0xdddd,0xd7d7,0x1d1d,0x1d1d,0x1d1d,0x2d2d,
 0x1d1d,0x1d1d,0x1d1d,0x1d1d,0xdddd,0xdddd,0x7d7d,0xdddd,0xdddd,0xdddd
#else
 0x6f6f,0x4f4f,0x1d1d,0xdede,0xdfdf,0x2e2e,0x7f7f,0x9e9e,0xaeae,0x7f7f,
 0x1d1d,0xdede,0xdfdf,0xdede,0xdfdf,0xdede,0xe1e1,0xdcdc,0x2e2e,0x1d1d,0xdcdc
#endif
};

/*
=====================
=
= DrawFlats
=
=====================
*/

void DrawFlats(unsigned tex_f, unsigned tex_c)
{
  int x, y, y0, halfheight, shade;
  unsigned top_offset0, bot_offset0, top_offset, bot_offset;
  unsigned top_add, bot_add;
  byte p, color;
  byte far *src_bot, far *src_top;
  fixed dist;            
  fixed tex_step;       
  fixed gu, gv, du, dv;
  int u, v;              

  halfheight=viewheight>>1;
  y0=min_wallheight>>3;      
  if(y0>halfheight) return;  
  if(y0==0) y0=1;           
  top_offset0=80*(halfheight-y0-1); 
  bot_offset0=80*(halfheight+y0);  

  src_bot=PM_GetPage(tex_f); 
  src_top=PM_GetPage(tex_c); 

  for(p=0; p<4; p++)
  {
    asm mov ax,0x0102
    asm mov cl,[p]
    asm shl ah,cl
    asm mov dx,0x3c4
    asm out dx,ax

    for(y=y0, top_offset=top_offset0, bot_offset=bot_offset0; y<halfheight; y++, top_offset-=80, bot_offset+=80)
    {
      dist=(heightnumerator/y)<<5;
      gu= viewx+FixedByFrac(dist, viewcos);
      gv=-viewy+FixedByFrac(dist, viewsin);
      tex_step=(dist<<8)/viewwidth/175;
      du= FixedByFrac(tex_step, viewsin);
      dv=-FixedByFrac(tex_step, viewcos);
      gu-=((viewwidth>>1)-p)*du;
      gv-=((viewwidth>>1)-p)*dv; 
      du<<=2; 
      dv<<=2;
      shade=(y<<2)/((maxscaleshl2>>8)+1+LSHADE_flag);
      if(shade>32) shade=32;
      shade=32-shade; 
      for(x=p, top_add=top_offset, bot_add=bot_offset; x<viewwidth; x+=4, top_add++, bot_add++)
      {
        if(wallheight[x]>>3<=y)
        {
          u=(gu>>10)&63; v=(gv>>10)&63;
          color=shadetable[shade][*(src_top+((63-u)<<6)+(63-v))]; 

          asm mov es,[screenseg]
          asm mov di,[bufferofs]
          asm add di,[top_add]
          asm mov al,[color]
          asm mov es:[di],al

          color=shadetable[shade][*(src_bot+(u<<6)+(63-v))];

          asm mov es,[screenseg]
          asm mov di,[bufferofs]
          asm add di,[bot_add]
          asm mov al,[color]
          asm mov es:[di],al
        }
        gu+=du; gv+=dv;
      }
    }
  }
} 

/*
=====================
=
= DrawTfloor
=
=====================
*/

void DrawTfloor(unsigned tex_f)
{
  int x, y, y0, halfheight, shade;
  unsigned bot_offset0, bot_offset;
  unsigned bot_add;
  byte p, color;
  byte far *src_bot;
  fixed dist;            
  fixed tex_step;       
  fixed gu, gv, du, dv;
  int u, v;              

  halfheight=viewheight>>1;
  y0=min_wallheight>>3;      
  if(y0>halfheight) return;  
  if(y0==0) y0=1;           
  bot_offset0=80*(halfheight+y0);  

  src_bot=PM_GetPage(tex_f);  

  for(p=0; p<4; p++)
  {
    asm mov ax,0x0102
    asm mov cl,[p]
    asm shl ah,cl
    asm mov dx,0x3c4
    asm out dx,ax

    for(y=y0, bot_offset=bot_offset0; y<halfheight; y++, bot_offset+=80)
    {
      dist=(heightnumerator/y)<<5;
      gu= viewx+FixedByFrac(dist, viewcos);
      gv=-viewy+FixedByFrac(dist, viewsin);
      tex_step=(dist<<8)/viewwidth/175;
      du= FixedByFrac(tex_step, viewsin);
      dv=-FixedByFrac(tex_step, viewcos);
      gu-=((viewwidth>>1)-p)*du;
      gv-=((viewwidth>>1)-p)*dv; 
      du<<=2; 
      dv<<=2;
      shade=(y<<2)/((maxscaleshl2>>8)+1+LSHADE_flag);
      if(shade>32) shade=32;
      shade=32-shade; 
      for(x=p, bot_add=bot_offset; x<viewwidth; x+=4, bot_add++)
      {
        if(wallheight[x]>>3<=y)
        {
          u=(gu>>10)&63; v=(gv>>10)&63;

          color=shadetable[shade][*(src_bot+(u<<6)+(63-v))];

          asm mov es,[screenseg]
          asm mov di,[bufferofs]
          asm add di,[bot_add]
          asm mov al,[color]
          asm mov es:[di],al
        }
        gu+=du; gv+=dv;
      }
    }
  }
} 

/*
=====================
=
= VGAClearScreen
=
=====================
*/


void VGAClearScreen (unsigned ceiling, unsigned floor, int SkyState)
{
word pitch=80; 
   if(!SkyState)
      { 
      asm{
         mov dx,SC_INDEX
         mov ax,SC_MAPMASK+15*256
         out dx,ax

         mov ax,[viewwidth]
         shr ax,2
         sub word ptr pitch,ax

         mov dx,[viewwidth]
         shr dx,3
         mov dh,BYTE PTR [viewheight]
         shr dh,1

         mov es,[screenseg]
         mov di,[bufferofs]

         xor ah,ah
         mov al,byte ptr [ceiling]
         push ds
         mov si,seg shadetable
         mov ds,si
         mov si,offset shadetable
         add si,ax

         mov al,byte ptr [si]
         mov ah,al

         mov bh,dh
         shr bh,5
         inc bh
         mov bl,bh

         xor ch,ch
         push dx
      }
toploop:
      asm{
         mov cl,dl
         mov al,byte ptr [floor] 
         rep stosw
         add di,word ptr pitch

         dec bl
         jz nexttshade

         dec dh
         jnz toploop
         jmp topfin
      }
nexttshade:
      asm{
         add si,256
         mov al,byte ptr [si]
         mov ah,al
         mov bl,bh
         dec dh
         jnz toploop
      }
topfin:
      asm{
         pop dx
         mov si,offset shadetable+(SHADE_COUNT-1)*256
         add si,0x19         // floor color
         mov al,byte ptr [si]
         mov ah,al
         mov bl,bh
      }
bottomloop:
      asm{
         mov cl,dl
         rep stosw
         add di,word ptr pitch

         dec bl
         jz nextbshade

         dec dh
         jnz bottomloop
         jmp botfin
      }
nextbshade:
      asm{
         sub si,256
         mov al,byte ptr [si]
         mov ah,al
         mov bl,bh
         dec dh
         jnz bottomloop
      }
botfin:
      asm{
         pop ds
      }
   }
   else
   {
      long siny,cosy;

      siny=(long)sintable[player->angle];
      cosy=(long)costable[player->angle];

      if(siny<0) siny=-(siny&0x7fffffff);
      if(cosy<0) cosy=-(cosy&0x7fffffff);

      PaintStars(siny,cosy);

      //
      // clear rest of the screen
      //
      asm{
         mov dx,SC_INDEX
         mov ax,SC_MAPMASK+15*256         // write through all planes
         out dx,ax

         mov es,[screenseg]
         mov di,[bufferofs]

         mov dx,80
         mov ax,[viewwidth]
         mov bx,ax
         shr ax,2
         sub dx,ax

         shr bx,3

         mov al,byte ptr [viewheight]
         mov bh,al
         shl ax,3
         add di,ax
         shl ax,2
         add di,ax

         shr bh,1

         mov ax,[floor]
         xor ch,ch
      }

bottomloop2:
      asm{
         mov cl,bl
         rep stosw
         add di,dx
         dec bh
         jnz bottomloop2
      }
   }
} 


//==========================================================================

/*
=====================
=
= CalcRotate
=
=====================
*/

int	CalcRotate (objtype *ob)
{
	int	angle,viewangle;

	// this isn't exactly correct, as it should vary by a trig value,
	// but it is close enough with only eight rotations

	viewangle = player->angle + (centerx - ob->viewx)/8;

	if (ob->obclass == rocketobj || ob->obclass == hrocketobj)
		angle =  (viewangle-180)- ob->angle;
	else
		angle =  (viewangle-180)- dirangle[ob->dir];

	angle+=ANGLES/16;
	while (angle>=ANGLES)
		angle-=ANGLES;
	while (angle<0)
		angle+=ANGLES;

	if (ob->state->rotate == 2)             // 2 rotation pain frame
		return 4*(angle/(ANGLES/2));        // seperated by 3 (art layout...)

	return angle/(ANGLES/8);
}


/*
=====================
=
= DrawScaleds
=
= Draws all objects that are visable
=
=====================
*/

#define MAXVISABLE	50

typedef struct
{
	int	viewx,
		viewheight,
		shapenum;
} visobj_t;

visobj_t	vislist[MAXVISABLE],*visptr,*visstep,*farthest;

void DrawScaleds (void)
{
	int 		i,j,least,numvisable,height;
	memptr		shape;
	byte		*tilespot,*visspot;
	int			shapenum;
	unsigned	spotloc;

	statobj_t	*statptr;
	objtype		*obj;

	visptr = &vislist[0];

//
// place static objects
//
	for (statptr = &statobjlist[0] ; statptr !=laststatobj ; statptr++)
	{
		if ((visptr->shapenum = statptr->shapenum) == -1)
			continue;						// object has been deleted

		if (!*statptr->visspot)
			continue;						// not visable

      		if (statptr->flags & FL_ANIMATED)
      		{
         		   visptr->shapenum+=(frameon>>3)%4;
      		}

		if (TransformTile (statptr->tilex,statptr->tiley
			,&visptr->viewx,&visptr->viewheight) && statptr->flags & FL_BONUS)
		{
			GetBonus (statptr);
			continue;
		}

		if (!visptr->viewheight)
			continue;						// to close to the object

		if (visptr < &vislist[MAXVISABLE-1])	// don't let it overflow
			visptr++;
	}

//
// place active objects
//
	for (obj = player->next;obj;obj=obj->next)
	{
		if (!(visptr->shapenum = obj->state->shapenum))
			continue;						// no shape

		spotloc = (obj->tilex<<6)+obj->tiley;	// optimize: keep in struct?
		visspot = &spotvis[0][0]+spotloc;
		tilespot = &tilemap[0][0]+spotloc;

		//
		// could be in any of the nine surrounding tiles
		//
		if (*visspot
		|| ( *(visspot-1) && !*(tilespot-1) )
		|| ( *(visspot+1) && !*(tilespot+1) )
		|| ( *(visspot-65) && !*(tilespot-65) )
		|| ( *(visspot-64) && !*(tilespot-64) )
		|| ( *(visspot-63) && !*(tilespot-63) )
		|| ( *(visspot+65) && !*(tilespot+65) )
		|| ( *(visspot+64) && !*(tilespot+64) )
		|| ( *(visspot+63) && !*(tilespot+63) ) )
		{
			obj->active = true;
			TransformActor (obj);
			if (!obj->viewheight)
				continue;						// too close or far away

			visptr->viewx = obj->viewx;
			visptr->viewheight = obj->viewheight;
			if (visptr->shapenum == -1)
				visptr->shapenum = obj->temp1;	// special shape

			if (obj->state->rotate)
				visptr->shapenum += CalcRotate (obj);

			if (visptr < &vislist[MAXVISABLE-1])	// don't let it overflow
				visptr++;
			obj->flags |= FL_VISABLE;
		}
		else
			obj->flags &= ~FL_VISABLE;
	}

//
// draw from back to front
//
	numvisable = visptr-&vislist[0];

	if (!numvisable)
		return;									// no visable objects

	for (i = 0; i<numvisable; i++)
	{
		least = 32000;
		for (visstep=&vislist[0] ; visstep<visptr ; visstep++)
		{
			height = visstep->viewheight;
			if (height < least)
			{
				least = height;
				farthest = visstep;
			}
		}
		//
		// draw farthest
		//
		ScaleShape(farthest->viewx,farthest->shapenum,farthest->viewheight);

		farthest->viewheight = 32000;
	}

}

//==========================================================================

/*
==============
=
= DrawPlayerWeapon
=
= Draw the player's hands
=
==============
*/

int	weaponscale[NUMWEAPONS] = {SPR_KNIFEREADY,SPR_PISTOLREADY
	,SPR_MACHINEGUNREADY,SPR_CHAINREADY,SPR_MCSGUNREADY
                ,SPR_MLAUNCHREADY};

void DrawPlayerWeapon (void)
{
	int	shapenum;

#ifndef SPEAR
	if (gamestate.victoryflag)
	{
		if (player->state == &s_deathcam && (TimeCount&32) )
			SimpleScaleShape(viewwidth/2,SPR_DEATHCAM,viewheight+1);
		return;
	}
#endif

	if (gamestate.weapon != -1)
	{
		shapenum = weaponscale[gamestate.weapon]+gamestate.weaponframe;
		SimpleScaleShape(viewwidth/2-10+gamestate.bobber2,shapenum,viewheight+1+gamestate.bobber);
	}

	if (demorecord || demoplayback)
		SimpleScaleShape(viewwidth/2,SPR_DEMO,viewheight+1);}


//==========================================================================


/*
=====================
=
= CalcTics
=
=====================
*/

void CalcTics (void)
{
	long	newtime,oldtimecount;

//
// calculate tics since last refresh for adaptive timing
//
	if (lasttimecount > TimeCount)
		TimeCount = lasttimecount;		// if the game was paused a LONG time

	do
	{
		newtime = TimeCount;
		tics = newtime-lasttimecount;
	} while (!tics);			// make sure at least one tic passes

	lasttimecount = newtime;

#ifdef FILEPROFILE
		strcpy (scratch,"\tTics:");
		itoa (tics,str,10);
		strcat (scratch,str);
		strcat (scratch,"\n");
		write (profilehandle,scratch,strlen(scratch));
#endif

	if (tics>MAXTICS)
	{
		TimeCount -= (tics-MAXTICS);
		tics = MAXTICS;
	}
}


//==========================================================================


/*
========================
=
= FixOfs
=
========================
*/

void	FixOfs (void)
{
	VW_ScreenToScreen (displayofs,bufferofs,viewwidth/8,viewheight);
}


//==========================================================================


/*
====================
=
= WallRefresh
=
====================
*/

void WallRefresh (void)
{
//
// set up variables for this view
//
	viewangle = player->angle;
	midangle = viewangle*(FINEANGLES/ANGLES);
	viewsin = sintable[viewangle];
	viewcos = costable[viewangle];
	viewx = player->x - FixedByFrac(focallength,viewcos);
	viewy = player->y + FixedByFrac(focallength,viewsin);

	focaltx = viewx>>TILESHIFT;
	focalty = viewy>>TILESHIFT;

	viewtx = player->x >> TILESHIFT;
	viewty = player->y >> TILESHIFT;

	xpartialdown = viewx&(TILEGLOBAL-1);
	xpartialup = TILEGLOBAL-xpartialdown;
	ypartialdown = viewy&(TILEGLOBAL-1);
	ypartialup = TILEGLOBAL-ypartialdown;

	lastside = -1;			// the first pixel is on a new wall
	AsmRefresh ();
	ScalePost ();			// no more optimization on last post
}

//==========================================================================

/*
========================
=
= ThreeDRefresh
=
========================
*/


void   ThreeDRefresh (void)
{
   int tracedir, texture, SkyState;
   unsigned ceiling, floor;
   boolean flats;
   boolean tfloor;

   ceiling=vgaCeiling[gamestate.episode*10+gamestate.mapon]; 
   floor = 0x1919;
   SkyState = 0;
   flats = false;
   tfloor = false;
   texture = 0;
   
      outportb (SC_INDEX,SC_MAPMASK);

   asm mov   ax,ds
   asm mov   es,ax
   asm mov   di,OFFSET spotvis
   asm xor   ax,ax
   asm mov   cx,2048   // 64*64 / 2
   asm rep stosw
   
   bufferofs += screenofs;

         switch (gamestate.mapon)
      {
      case 0:
         flats = true;
         tfloor = false;
         texture = 52;
         break;
      case 1:
         flats = false;
         tfloor = true;
         texture = 2;
         SkyState = 3;
         break;
      case 2:
         flats = true;
         tfloor = false;
         texture = 52;
         break;
      case 3:
         flats = false;
         tfloor = true;
         texture = 3;
         SkyState = 2;
         break;
      case 4:
         flats = true;
         tfloor = false;
         texture = 52;
         break;
      case 5:
         flats = false;
         tfloor = true;
         texture = 2;
         SkyState = 3;
         break;
      case 6:
         flats = true;
         tfloor = false;
         texture = 58;
         break; 
      case 7:
         flats = true;
         tfloor = false;
         texture = 58;
         break;
      case 8:
         flats = true;
         tfloor = false;
         texture = 58;
         break;
      case 9:
         flats = true;
         tfloor = false;
         texture = 58;
         break;
      case 10:
         flats = true;
         tfloor = false;
         texture = 60;
         break;
      case 11:
         flats = true;
         tfloor = false;
         texture = 60;
         break;
      case 12:
         flats = false;
         tfloor = true;
         texture = 3;
         SkyState = 2;
         break;
      case 13:
         flats = true;
         tfloor = false;
         texture = 60;
         break;
      case 14:
         flats = false;
         tfloor = true;
         texture = 3;
         SkyState = 1;
         break;
      case 15:
         flats = true;
         tfloor = false;
         texture = 62;
         break;
      case 16:
         flats = false;
         tfloor = true;
         texture = 3;
         SkyState = 1;
         break;
      case 17:
         flats = true;
         tfloor = false;
         texture = 62;
         break;
      case 18:
         flats = true;
         tfloor = false;
         texture = 54;
         break;
      case 19:
         flats = true;
         tfloor = false;
         texture = 54;
         break;
      case 20:
         flats = true;
         tfloor = false;
         texture = 54;
         break;
      }
      
      if (!flats) VGAClearScreen(ceiling, floor, SkyState);
      
      WallRefresh();
      
      if (flats) DrawFlats(texture, texture + 1);
      if (tfloor) DrawTfloor(texture);

      DrawScaleds();

      if(SkyState==3) PaintSnow(viewsin,viewcos);
      else if(SkyState==2) PaintRain(viewsin,viewcos);

        DrawPlayerWeapon ();   

      if (fizzlein)
         {
            FizzleFade(bufferofs,displayofs+screenofs,viewwidth,viewheight,20,false);
            fizzlein = false;
            lasttimecount = TimeCount = 0; // don't make a big tic count
         }

      bufferofs -= screenofs;
      displayofs = bufferofs;

      asm cli
      asm mov   cx,[displayofs]
      asm mov   dx,3d4h
      asm mov   al,0ch
      asm out   dx,al
      asm inc   dx
      asm mov   al,ch
      asm out   dx,al
      asm sti
      
      
      bufferofs += SCREENSIZE;
      if (bufferofs > PAGE3START) bufferofs = PAGE1START;

      frameon++;
      PM_NextFrame();
}

//===========================================================================